@@ -4,18 +4,17 @@ import optparse
 
 import sys
 import re
-import difflib
 
-from pip.basecommand import command_dict, load_command, load_all_commands, command_names
-from pip.baseparser import parser
-from pip.exceptions import InstallationError
+from pip.exceptions import InstallationError, CommandError, PipError
 from pip.log import logger
 from pip.util import get_installed_distributions, get_prog
-from pip.vcs import git, mercurial, subversion, bazaar
+from pip.vcs import git, mercurial, subversion, bazaar  # noqa
+from pip.baseparser import create_main_parser
+from pip.commands import commands, get_similar_commands, get_summaries
 
 
 # The version as used in the setup.py and the docs conf.py
-__version__ = "1.2"
+__version__ = "1.3"
 
 
 def autocomplete():
@@ -33,14 +32,16 @@ def autocomplete():
         current = cwords[cword - 1]
     except IndexError:
         current = ''
-    load_all_commands()
-    subcommands = [cmd for cmd, cls in command_dict.items() if not cls.hidden]
+
+    subcommands = [cmd for cmd, summary in get_summaries()]
     options = []
     # subcommand
     try:
         subcommand_name = [w for w in cwords if w in subcommands][0]
     except IndexError:
         subcommand_name = None
+
+    parser = create_main_parser()
     # subcommand options
     if subcommand_name:
         # special case: 'help' subcommand has no options
@@ -58,10 +59,12 @@ def autocomplete():
                 for dist in installed:
                     print(dist)
                 sys.exit(1)
-        subcommand = command_dict.get(subcommand_name)
+
+        subcommand = commands[subcommand_name](parser)
         options += [(opt.get_opt_string(), opt.nargs)
-                    for opt in subcommand.parser.option_list
+                    for opt in subcommand.parser.option_list_all
                     if opt.help != optparse.SUPPRESS_HELP]
+
         # filter out previously specified options from available options
         prev_opts = [x.split('=')[0] for x in cwords[1:cword - 1]]
         options = [(x, v) for (x, v) in options if x not in prev_opts]
@@ -74,40 +77,75 @@ def autocomplete():
                 opt_label += '='
             print(opt_label)
     else:
-        # show options of main parser only when necessary
+        # show main parser options only when necessary
         if current.startswith('-') or current.startswith('--'):
-            subcommands += [opt.get_opt_string()
-                            for opt in parser.option_list
-                            if opt.help != optparse.SUPPRESS_HELP]
+            opts = [i.option_list for i in parser.option_groups]
+            opts.append(parser.option_list)
+            opts = (o for it in opts for o in it)
+
+            subcommands += [i.get_opt_string() for i in opts
+                            if i.help != optparse.SUPPRESS_HELP]
+
         print(' '.join([x for x in subcommands if x.startswith(current)]))
     sys.exit(1)
 
 
+def parseopts(args):
+    parser = create_main_parser()
+    parser.main = True # so the help formatter knows
+
+    # create command listing
+    command_summaries = get_summaries()
+
+    description = [''] + ['%-27s %s' % (i, j) for i, j in command_summaries]
+    parser.description = '\n'.join(description)
+
+    options, args = parser.parse_args(args)
+
+    if options.version:
+        sys.stdout.write(parser.version)
+        sys.stdout.write(os.linesep)
+        sys.exit()
+
+    # pip || pip help || pip --help -> print_help()
+    if not args or (args[0] == 'help' and len(args) == 1):
+        parser.print_help()
+        sys.exit()
+
+    if not args:
+        msg = ('You must give a command '
+               '(use "pip --help" to see a list of commands)')
+        raise CommandError(msg)
+
+    command = args[0].lower()
+
+    if command not in commands:
+        guess = get_similar_commands(command)
+
+        msg = ['unknown command "%s"' % command]
+        if guess:
+            msg.append('maybe you meant "%s"' % guess)
+
+        raise CommandError(' - '.join(msg))
+
+    return command, options, args, parser
+
+
 def main(initial_args=None):
     if initial_args is None:
         initial_args = sys.argv[1:]
+
     autocomplete()
-    options, args = parser.parse_args(initial_args)
-    if options.help and not args:
-        args = ['help']
-    if not args:
-        parser.error('You must give a command '
-            '(use "%s help" to see a list of commands)' % get_prog())
-    command = args[0].lower()
-    load_command(command)
-    if command not in command_dict:
-        close_commands = difflib.get_close_matches(command, command_names())
-        if close_commands:
-            guess = close_commands[0]
-            if args[1:]:
-                guess = "%s %s" % (guess, " ".join(args[1:]))
-        else:
-            guess = 'install %s' % command
-        error_dict = {'arg': command, 'guess': guess,
-                      'script': os.path.basename(sys.argv[0])}
-        parser.error('No command by the name %(script)s %(arg)s\n  '
-                     '(maybe you meant "%(script)s %(guess)s")' % error_dict)
-    command = command_dict[command]
+
+    try:
+        cmd_name, options, args, parser = parseopts(initial_args)
+    except PipError:
+        e = sys.exc_info()[1]
+        sys.stderr.write("ERROR: %s" % e)
+        sys.stderr.write(os.linesep)
+        sys.exit(1)
+
+    command = commands[cmd_name](parser)  # see baseparser.Command
     return command.main(args[1:], options)
 
 
