@@ -3,29 +3,36 @@
 import sys
 import os
 import re
+import gzip
 import mimetypes
-import threading
 import posixpath
 import pkg_resources
-import urllib
-import urllib2
-import urlparse
 import random
 import socket
 import string
-from Queue import Queue
-from Queue import Empty as QueueEmpty
+import zlib
+
+try:
+    import threading
+except ImportError:
+    import dummy_threading as threading
+
 from pip.log import logger
-from pip.util import Inf
-from pip.util import normalize_name, splitext
-from pip.exceptions import DistributionNotFound
-from pip.backwardcompat import WindowsError, product
+from pip.util import Inf, normalize_name, splitext
+from pip.exceptions import DistributionNotFound, BestVersionAlreadyInstalled
+from pip.backwardcompat import (WindowsError, BytesIO,
+                                Queue, urlparse,
+                                URLError, HTTPError, u,
+                                product, url2pathname, ssl,
+                                Empty as QueueEmpty)
+if ssl:
+    from pip.backwardcompat import CertificateError
 from pip.download import urlopen, path_to_url2, url_to_path, geturl, Urllib2HeadRequest
 
 __all__ = ['PackageFinder']
 
 
-DEFAULT_MIRROR_URL = "last.pypi.python.org"
+DEFAULT_MIRROR_HOSTNAME = "last.pypi.python.org"
 
 
 class PackageFinder(object):
@@ -56,8 +63,7 @@ class PackageFinder(object):
         ## FIXME: also, we should track comes_from (i.e., use Link)
         self.dependency_links.extend(links)
 
-    @staticmethod
-    def _sort_locations(locations):
+    def _sort_locations(self, locations):
         """
         Sort locations into "files" (archives) and "urls", and return
         a pair of lists (files,urls)
@@ -65,8 +71,7 @@ class PackageFinder(object):
         files = []
         urls = []
 
-        # puts the url for the given file path into the appropriate
-        # list
+        # puts the url for the given file path into the appropriate list
         def sort_path(path):
             url = path_to_url2(path)
             if mimetypes.guess_type(url, strict=False)[0] == 'text/html':
@@ -75,33 +80,30 @@ class PackageFinder(object):
                 files.append(url)
 
         for url in locations:
-            if url.startswith('file:'):
-                path = url_to_path(url)
-                if os.path.isdir(path):
+
+            is_local_path = os.path.exists(url)
+            is_file_url = url.startswith('file:')
+            is_find_link = url in self.find_links
+
+            if is_local_path or is_file_url:
+                if is_local_path:
+                    path = url
+                else:
+                    path = url_to_path(url)
+                if is_find_link and os.path.isdir(path):
                     path = os.path.realpath(path)
                     for item in os.listdir(path):
                         sort_path(os.path.join(path, item))
+                elif is_file_url and os.path.isdir(path):
+                    urls.append(url)
                 elif os.path.isfile(path):
                     sort_path(path)
             else:
                 urls.append(url)
+
         return files, urls
 
     def find_requirement(self, req, upgrade):
-        url_name = req.url_name
-        # Only check main index if index URL is given:
-        main_index_url = None
-        if self.index_urls:
-            # Check that we have the url_name correctly spelled:
-            main_index_url = Link(posixpath.join(self.index_urls[0], url_name))
-            # This will also cache the page, so it's okay that we get it again later:
-            page = self._get_page(main_index_url, req)
-            if page is None:
-                url_name = self._find_url_name(Link(self.index_urls[0]), url_name, req) or req.url_name
-
-        # Combine index URLs with mirror URLs here to allow
-        # adding more index URLs from requirements files
-        all_index_urls = self.index_urls + self.mirror_urls
 
         def mkurl_pypi_url(url):
             loc = posixpath.join(url, url_name)
@@ -112,6 +114,22 @@ class PackageFinder(object):
             if not loc.endswith('/'):
                 loc = loc + '/'
             return loc
+
+        url_name = req.url_name
+        # Only check main index if index URL is given:
+        main_index_url = None
+        if self.index_urls:
+            # Check that we have the url_name correctly spelled:
+            main_index_url = Link(mkurl_pypi_url(self.index_urls[0]))
+            # This will also cache the page, so it's okay that we get it again later:
+            page = self._get_page(main_index_url, req)
+            if page is None:
+                url_name = self._find_url_name(Link(self.index_urls[0]), url_name, req) or req.url_name
+
+        # Combine index URLs with mirror URLs here to allow
+        # adding more index URLs from requirements files
+        all_index_urls = self.index_urls + self.mirror_urls
+
         if url_name is not None:
             locations = [
                 mkurl_pypi_url(url)
@@ -151,43 +169,46 @@ class PackageFinder(object):
         if not found_versions and not page_versions and not dependency_versions and not file_versions:
             logger.fatal('Could not find any downloads that satisfy the requirement %s' % req)
             raise DistributionNotFound('No distributions at all found for %s' % req)
+        installed_version = []
         if req.satisfied_by is not None:
-            found_versions.append((req.satisfied_by.parsed_version, Inf, req.satisfied_by.version))
+            installed_version = [(req.satisfied_by.parsed_version, InfLink, req.satisfied_by.version)]
         if file_versions:
             file_versions.sort(reverse=True)
             logger.info('Local files found: %s' % ', '.join([url_to_path(link.url) for parsed, link, version in file_versions]))
-            found_versions = file_versions + found_versions
-        all_versions = found_versions + page_versions + dependency_versions
+        #this is an intentional priority ordering
+        all_versions = installed_version + file_versions + found_versions + page_versions + dependency_versions
         applicable_versions = []
         for (parsed_version, link, version) in all_versions:
             if version not in req.req:
                 logger.info("Ignoring link %s, version %s doesn't match %s"
                             % (link, version, ','.join([''.join(s) for s in req.req.specs])))
                 continue
-            applicable_versions.append((link, version))
-        applicable_versions = sorted(applicable_versions, key=lambda v: pkg_resources.parse_version(v[1]), reverse=True)
-        existing_applicable = bool([link for link, version in applicable_versions if link is Inf])
+            applicable_versions.append((parsed_version, link, version))
+        #bring the latest version to the front, but maintains the priority ordering as secondary
+        applicable_versions = sorted(applicable_versions, key=lambda v: v[0], reverse=True)
+        existing_applicable = bool([link for parsed_version, link, version in applicable_versions if link is InfLink])
         if not upgrade and existing_applicable:
-            if applicable_versions[0][1] is Inf:
+            if applicable_versions[0][1] is InfLink:
                 logger.info('Existing installed version (%s) is most up-to-date and satisfies requirement'
                             % req.satisfied_by.version)
             else:
                 logger.info('Existing installed version (%s) satisfies requirement (most up-to-date version is %s)'
-                            % (req.satisfied_by.version, applicable_versions[0][1]))
+                            % (req.satisfied_by.version, applicable_versions[0][2]))
             return None
         if not applicable_versions:
             logger.fatal('Could not find a version that satisfies the requirement %s (from versions: %s)'
-                         % (req, ', '.join([version for parsed_version, link, version in found_versions])))
+                         % (req, ', '.join([version for parsed_version, link, version in all_versions])))
             raise DistributionNotFound('No distributions matching the version for %s' % req)
-        if applicable_versions[0][0] is Inf:
+        if applicable_versions[0][1] is InfLink:
             # We have an existing version, and its the best version
             logger.info('Installed version (%s) is most up-to-date (past versions: %s)'
-                        % (req.satisfied_by.version, ', '.join([version for link, version in applicable_versions[1:]]) or 'none'))
-            return None
+                        % (req.satisfied_by.version, ', '.join([version for parsed_version, link, version in applicable_versions[1:]]) or 'none'))
+            raise BestVersionAlreadyInstalled
         if len(applicable_versions) > 1:
             logger.info('Using version %s (newest of versions: %s)' %
-                        (applicable_versions[0][1], ', '.join([version for link, version in applicable_versions])))
-        return applicable_versions[0][0]
+                        (applicable_versions[0][2], ', '.join([version for parsed_version, link, version in applicable_versions])))
+        return applicable_versions[0][1]
+
 
     def _find_url_name(self, index_url, url_name, req):
         """Finds the true URL name of a package, when the given name isn't quite correct.
@@ -210,7 +231,7 @@ class PackageFinder(object):
 
     def _get_pages(self, locations, req):
         """Yields (page, page_url) from the given locations, skipping
-        locations that have errors, and adding download/homepage links"""
+        locations that have errors, and adding homepage links"""
         pending_queue = Queue()
         for location in locations:
             pending_queue.put(location)
@@ -241,12 +262,12 @@ class PackageFinder(object):
             if page is None:
                 continue
             done.append(page)
-            for link in page.rel_links():
+            for link in page.rel_links(rels=('homepage',)):
                 pending_queue.put(link)
 
     _egg_fragment_re = re.compile(r'#egg=([^&]*)')
     _egg_info_re = re.compile(r'([a-z0-9_.]+)-([a-z0-9_.-]+)', re.I)
-    _py_version_re = re.compile(r'-py([123]\.[0-9])$')
+    _py_version_re = re.compile(r'-py([123]\.?[0-9]?)$')
 
     def _sort_links(self, links):
         "Returns elements of links in order, non-egg links first, egg links second, while eliminating duplicates"
@@ -292,6 +313,11 @@ class PackageFinder(object):
                     logger.debug('Skipping link %s; unknown archive format: %s' % (link, ext))
                     self.logged_links.add(link)
                 return []
+            if "macosx10" in link.path and ext == '.zip':
+                if link not in self.logged_links:
+                    logger.debug('Skipping link %s; macosx10 one' % (link))
+                    self.logged_links.add(link)
+                return []
         version = self._egg_info_matches(egg_info, search_name, link)
         if version is None:
             logger.debug('Skipping link %s; wrong project name (not %s)' % (link, search_name))
@@ -316,8 +342,10 @@ class PackageFinder(object):
         name = match.group(0).lower()
         # To match the "safe" name that pkg_resources creates:
         name = name.replace('_', '-')
-        if name.startswith(search_name.lower()):
-            return match.group(0)[len(search_name):].lstrip('-')
+        # project name and version must be separated by a dash
+        look_for = search_name.lower() + "-"
+        if name.startswith(look_for):
+            return match.group(0)[len(look_for):]
         else:
             return None
 
@@ -335,12 +363,13 @@ class PackageFinder(object):
 
         mirror_urls = set()
         for mirror_url in mirrors:
+            mirror_url = mirror_url.rstrip('/')
             # Make sure we have a valid URL
-            if not ("http://" or "https://" or "file://") in mirror_url:
+            if not any([mirror_url.startswith(scheme) for scheme in ["http://", "https://", "file://"]]):
                 mirror_url = "http://%s" % mirror_url
             if not mirror_url.endswith("/simple"):
-                mirror_url = "%s/simple/" % mirror_url
-            mirror_urls.add(mirror_url)
+                mirror_url = "%s/simple" % mirror_url
+            mirror_urls.add(mirror_url + '/')
 
         return list(mirror_urls)
 
@@ -432,7 +461,7 @@ class HTMLPage(object):
 
             # Tack index.html onto file:// URLs that point to directories
             (scheme, netloc, path, params, query, fragment) = urlparse.urlparse(url)
-            if scheme == 'file' and os.path.isdir(urllib.url2pathname(path)):
+            if scheme == 'file' and os.path.isdir(url2pathname(path)):
                 # add trailing slash if not present so urljoin doesn't trim final segment
                 if not url.endswith('/'):
                     url += '/'
@@ -443,21 +472,35 @@ class HTMLPage(object):
 
             real_url = geturl(resp)
             headers = resp.info()
-            inst = cls(resp.read(), real_url, headers)
-        except (urllib2.HTTPError, urllib2.URLError, socket.timeout, socket.error, OSError, WindowsError), e:
+            contents = resp.read()
+            encoding = headers.get('Content-Encoding', None)
+            #XXX need to handle exceptions and add testing for this
+            if encoding is not None:
+                if encoding == 'gzip':
+                    contents = gzip.GzipFile(fileobj=BytesIO(contents)).read()
+                if encoding == 'deflate':
+                    contents = zlib.decompress(contents)
+            inst = cls(u(contents), real_url, headers)
+        except (HTTPError, URLError, socket.timeout, socket.error, OSError, WindowsError):
+            e = sys.exc_info()[1]
             desc = str(e)
             if isinstance(e, socket.timeout):
                 log_meth = logger.info
                 level =1
                 desc = 'timed out'
-            elif isinstance(e, urllib2.URLError):
-                log_meth = logger.info
+            elif isinstance(e, URLError):
+                #ssl/certificate error
+                if ssl and hasattr(e, 'reason') and (isinstance(e.reason, ssl.SSLError) or isinstance(e.reason, CertificateError)):
+                    desc = 'There was a problem confirming the ssl certificate: %s' % e
+                    log_meth = logger.notify
+                else:
+                    log_meth = logger.info
                 if hasattr(e, 'reason') and isinstance(e.reason, socket.timeout):
                     desc = 'timed out'
                     level = 1
                 else:
                     level = 2
-            elif isinstance(e, urllib2.HTTPError) and e.code == 404:
+            elif isinstance(e, HTTPError) and e.code == 404:
                 ## FIXME: notify?
                 log_meth = logger.info
                 level = 2
@@ -509,8 +552,8 @@ class HTMLPage(object):
             url = self.clean_link(urlparse.urljoin(self.base_url, url))
             yield Link(url, self)
 
-    def rel_links(self):
-        for url in self.explicit_rel_links():
+    def rel_links(self, rels=('homepage', 'download')):
+        for url in self.explicit_rel_links(rels):
             yield url
         for url in self.scraped_rel_links():
             yield url
@@ -539,7 +582,7 @@ class HTMLPage(object):
             href_match = self._href_re.search(self.content, pos=match.end())
             if not href_match:
                 continue
-            url = match.group(1) or match.group(2) or match.group(3)
+            url = href_match.group(1) or href_match.group(2) or href_match.group(3)
             if not url:
                 continue
             url = self.clean_link(urlparse.urljoin(self.base_url, url))
@@ -565,7 +608,7 @@ class Link(object):
         if self.comes_from:
             return '%s (from %s)' % (self.url, self.comes_from)
         else:
-            return self.url
+            return str(self.url)
 
     def __repr__(self):
         return '<Link %s>' % self
@@ -573,18 +616,29 @@ class Link(object):
     def __eq__(self, other):
         return self.url == other.url
 
+    def __ne__(self, other):
+        return self.url != other.url
+
+    def __lt__(self, other):
+        return self.url < other.url
+
+    def __le__(self, other):
+        return self.url <= other.url
+
+    def __gt__(self, other):
+        return self.url > other.url
+
+    def __ge__(self, other):
+        return self.url >= other.url
+
     def __hash__(self):
         return hash(self.url)
 
     @property
     def filename(self):
-        url = self.url
-        url = url.split('#', 1)[0]
-        url = url.split('?', 1)[0]
-        url = url.rstrip('/')
-        name = posixpath.basename(url)
-        assert name, (
-            'URL %r produced no filename' % url)
+        _, netloc, path, _, _ = urlparse.urlsplit(self.url)
+        name = posixpath.basename(path.rstrip('/')) or netloc
+        assert name, ('URL %r produced no filename' % self.url)
         return name
 
     @property
@@ -598,6 +652,11 @@ class Link(object):
     def splitext(self):
         return splitext(posixpath.basename(self.path.rstrip('/')))
 
+    @property
+    def url_without_fragment(self):
+        scheme, netloc, path, query, fragment = urlparse.urlsplit(self.url)
+        return urlparse.urlunsplit((scheme, netloc, path, query, None))
+
     _egg_fragment_re = re.compile(r'#egg=([^&]*)')
 
     @property
@@ -607,11 +666,18 @@ class Link(object):
             return None
         return match.group(1)
 
-    _md5_re = re.compile(r'md5=([a-f0-9]+)')
+    _hash_re = re.compile(r'(sha1|sha224|sha384|sha256|sha512|md5)=([a-f0-9]+)')
 
     @property
-    def md5_hash(self):
-        match = self._md5_re.search(self.url)
+    def hash(self):
+        match = self._hash_re.search(self.url)
+        if match:
+            return match.group(2)
+        return None
+
+    @property
+    def hash_name(self):
+        match = self._hash_re.search(self.url)
         if match:
             return match.group(1)
         return None
@@ -620,6 +686,9 @@ class Link(object):
     def show_url(self):
         return posixpath.basename(self.url.split('#', 1)[0].split('?', 1)[0])
 
+#An "Infinite Link" that compares greater than other links
+InfLink = Link(Inf) #this object is not currently used as a sortable
+
 
 def get_requirement_from_url(url):
     """Get a requirement from the URL, if possible.  This looks for #egg
@@ -658,14 +727,17 @@ def get_mirrors(hostname=None):
     Originally written for the distutils2 project by Alexis Metaireau.
     """
     if hostname is None:
-        hostname = DEFAULT_MIRROR_URL
+        hostname = DEFAULT_MIRROR_HOSTNAME
 
     # return the last mirror registered on PyPI.
+    last_mirror_hostname = None
     try:
-        hostname = socket.gethostbyname_ex(hostname)[0]
+        last_mirror_hostname = socket.gethostbyname_ex(hostname)[0]
     except socket.gaierror:
         return []
-    end_letter = hostname.split(".", 1)
+    if not last_mirror_hostname or last_mirror_hostname == DEFAULT_MIRROR_HOSTNAME:
+        last_mirror_hostname = "z.pypi.python.org"
+    end_letter = last_mirror_hostname.split(".", 1)
 
     # determine the list from the last one.
     return ["%s.%s" % (s, end_letter[1]) for s in string_range(end_letter[0])]
